﻿-- Author      : RisM
-- Create Date : 6/28/2009 3:57:34 PM

SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L			= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)
local HELPFILE	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_HELPFILE",			false)
local DEFAULT_SPEECHES	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell_DEFAULT_SPEECHES",	false)

-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - GENERAL SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:GeneralOptions_OnShowSpellEventsToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowSpellEvents
	else -- "SET"
		SpeakinSpellSavedData.ShowSpellEvents = value
	end
end


function SpeakinSpell:GeneralOptions_OnShowDebugMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.ShowDebugMessages
	else -- "SET"
		SpeakinSpellSavedData.ShowDebugMessages = value
	end
end


function SpeakinSpell:GeneralOptions_OnEnableAllMessagesToggle(getset, value)
	if "GET" == getset then
		return SpeakinSpellSavedData.EnableAllMessages
	else -- "SET"
		SpeakinSpellSavedData.EnableAllMessages = value
	end
end


function SpeakinSpell:GeneralOptions_OnClickImport()
	self:ImportNewDefaults()
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - MESSAGE SETTINGS
-------------------------------------------------------------------------------


function SpeakinSpell:CurrentMessagesGUI_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if eventkey then
		return SpeakinSpellSavedData.EventTable[eventkey]
	else
		return nil
	end
end

function SpeakinSpell:CurrentMessagesGUI_ValidateSelectedFilter()
	-- if nothing is selected, select "show all"
	if not self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter then
		self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter = "ALL"
	end
	
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end

	

function SpeakinSpell:CurrentMessagesGUI_ValidateSelectedEvent()
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		if self:MatchesFilter(EventTableEntry.DetectedEvent, self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter) then
			-- the current selection is valid, so keep it
			return
		end
		--else, the current selection did not match the filter, so select a default
	end
	
	-- select the first thing in the list, if possible
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		if self:MatchesFilter(EventTableEntry.DetectedEvent, self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter) then
			-- the current selection is valid, so keep it
			self.OptionsGUIStates.MessageSettings.SelectedEventKey = key
			return
		end
	end
	
	-- our spell list is empty or nothing matches the filter
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
end



function SpeakinSpell:CurrentMessagesGUI_OnClickDelete()
	local funcname = "CurrentMessagesGUI_OnClickDelete"

	-- get the selected EventTableEntry
	local eventkey = self.OptionsGUIStates.MessageSettings.SelectedEventKey
	if not eventkey then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- delete the current selection from the SavedData table
	self:DeleteSpell(eventkey)
end


function SpeakinSpell:CurrentMessagesGUI_OnClickGotoCreate()
	InterfaceOptionsFrame_OpenToCategory(L["Create New..."])
end


function SpeakinSpell:CurrentMessagesGUI_OnEventTypeFilterSelect(getset, value)
	if "GET" == getset then
		--if nothing is selected, set to all
		self:CurrentMessagesGUI_ValidateSelectedFilter()
		return self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		local rebuild = not (self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter == value)
		-- set the new selection
		self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter = value
		-- rebuild the list to match the new filter
		if rebuild then
			self:CurrentMessagesGUI_RebuildSpellList()
		end
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end


function SpeakinSpell:CurrentMessagesGUI_OnEventSelect(getset, value)
	if "GET" == getset then
		-- create a side-effect to update the shown/hidden state of all options GUI controls
		self:CurrentMessagesGUI_Refresh()
		-- return the current spell selection
		return self.OptionsGUIStates.MessageSettings.SelectedEventKey
	else -- "SET"
		-- update the current selection
		self.OptionsGUIStates.MessageSettings.SelectedEventKey = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnChannelSelect(getset, scenario, value)
	local funcname = "CurrentMessagesGUI_OnChannelSelect"
		
	-- get the selected EventTableEntry
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg(funcname, "no event selected")
		return
	end

	if "GET" == getset then
		--self:DebugMsg(funcname,"GET "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
		if EventTableEntry.Channels[scenario] then
			return L[EventTableEntry.Channels[scenario]]
		else
			return L["Silent"]
		end
	else --"SET"
		-- store the new channel selection
		if value == L["Silent"] then
			EventTableEntry.Channels[scenario] = nil
		else
			-- create a reverse lookup table to un-localize channel names
			-- so we can return a channel name that is compatible with SendChatMessage()
			local CHANNEL_TABLE = {
				[L["SAY"]]			= "SAY",
				[L["PARTY"]]		= "PARTY",
				[L["RAID"]]			= "RAID",
				[L["EMOTE"]]		= "EMOTE",
				[L["BATTLEGROUND"]]	= "BATTLEGROUND",
				[L["YELL"]]			= "YELL",
				[L["RAID_WARNING"]]	= "RAID_WARNING",
			}
			EventTableEntry.Channels[scenario] = CHANNEL_TABLE[value]
		end
		--self:DebugMsg(funcname,"set "..scenario..":"..tostring(EventTableEntry.Channels[scenario]))
	end
end



function SpeakinSpell:CurrentMessagesGUI_OnWhisperTarget(getset, value)
	-- make sure we have a valid spell selection
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg( "CurrentMessagesGUI_OnWhisperTarget", "no event selected" )
		return false
	end
	
	if "GET" == getset then
		return EventTableEntry.WhisperTarget
	else
		EventTableEntry.WhisperTarget = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnMessage(getset, index, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnMessage", "no event selected")
		return ""
	end
	
	if "GET" == getset then
		local msg = EventTableEntry.Messages[index]
		if not msg then
			msg = ""
		end
		return msg
	else -- "SET"
		-- make sure we have a message table for the currently selected spell
		if not EventTableEntry.Messages then
			EventTableEntry.Messages = {}
		end
		
		-- store the new value into the message table
		EventTableEntry.Messages[index] = value
		self:DebugMsg( nil,"Message " .. tostring(index) .. " for spell " .. EventTableEntry.DetectedEvent.DisplayName .. " set to: " .. tostring(EventTableEntry.Messages[index]) )
		
		-- remove holes from the message table
		-- if the user has just deleted one of the entries, move everything else up to fill in the gap
		if value == "" then
			local Messages = EventTableEntry.Messages -- create reference for shortcut notation
			for i=index+1,SpeakinSpell.MAX.MESSAGES_PER_SPELL,1 do
				Messages[i-1] = Messages[i]
			end
			Messages[SpeakinSpell.MAX.MESSAGES_PER_SPELL] = nil
			-- destroy empty strings from the table
			for i=SpeakinSpell.MAX.MESSAGES_PER_SPELL-1,1,-1 do
				if Messages[i] == "" then
					Messages[i] = nil
				else
					break --the rest should be valid non-empty strings
				end
			end
			-- hide the superfluous input boxes
			self:CurrentMessagesGUI_Refresh()
		end
	end
end




function SpeakinSpell:CurrentMessagesGUI_OnSpellDisable(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnSpellDisable", "no event selected")
		return
	end

	if "GET" == getset then
		return EventTableEntry.DisableAnnouncements
	else -- "SET"
		EventTableEntry.DisableAnnouncements = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnFrequencySelect(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnFrequencySelect", "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Frequency
	else --"SET"
		EventTableEntry.Frequency = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_OnCooldownSlider(getset, value)
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if not EventTableEntry then
		self:ErrorMsg("CurrentMessagesGUI_OnCooldownSlider", "no event selected")
		return
	end
	
	if "GET" == getset then	
		return EventTableEntry.Cooldown
	else --"SET"
		EventTableEntry.Cooldown = value
	end
end


function SpeakinSpell:CurrentMessagesGUI_Refresh()
	local funcname = "CurrentMessagesGUI_Refresh"
	
	-- safety check in case this function is called early
	if not self.OptionsGUI then
		self:DebugMsg(funcname,"called prematurely, abort")
		return
	end
	
	-- determine how many Spell Message N input boxes are shown
	-- typically we'll show at least 1 speech message input box if we have a spell selected
	-- we want to show 1 more than the number of speeches currently configured
	-- unless we don't have a spell selected or we're in EnterNewSpell mode
	local NumSpellMessageInputsShown = 0
	local EventTableEntry = self:CurrentMessagesGUI_GetSelectedEventObject()
	if EventTableEntry then
		NumSpellMessageInputsShown = 1 + #(EventTableEntry.Messages)
	end
	self:DebugMsg(funcname,"NumSpellMessageInputsShown:"..tostring(NumSpellMessageInputsShown))
	
	-- show SpellMessageN input box if N <= NumSpellMessageInputsShown (numbering starts at 1)
	for i=1,SpeakinSpell.MAX.MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.CurrentMessagesGUI.args["SpellMessage"..i].guiHidden = (i > NumSpellMessageInputsShown)
	end
	
	-- hide all the options if a spell is not selected
	local HideControls = false
	if not EventTableEntry then
		HideControls = true
	end
	
	self.OptionsGUI.args.CurrentMessagesGUI.args["DeleteSpellButton"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["Disable"				].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectCaption"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectBG"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectRaid"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectParty"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectSolo"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["ChannelSelectArena"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["WhisperTargetToggle"	].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["FrequencySelect"		].guiHidden = HideControls
	self.OptionsGUI.args.CurrentMessagesGUI.args["CooldownSlider"		].guiHidden = HideControls
end



function SpeakinSpell:CurrentMessagesGUI_RebuildSpellList()
	-- apply the event filter while rebuilding the GUI list
	self:CurrentMessagesGUI_ValidateSelectedFilter()
	
	-- reset the spell list
	self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values = {}
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = nil
	
	-- search all entries in the event table for events that match the current filter
	for key,EventTableEntry in pairs(SpeakinSpellSavedData.EventTable) do
		if self:MatchesFilter(EventTableEntry.DetectedEvent, self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter) then
			-- add to the list of shown values
			self.OptionsGUI.args.CurrentMessagesGUI.args.EventSelect.values[key] = EventTableEntry.DetectedEvent.DisplayName
			-- make sure we have a current spell selected for the options GUI 
			-- we'll use the last one we see because it's easier
			self.OptionsGUIStates.MessageSettings.SelectedEventKey = key
		end
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - CREATE NEW
-------------------------------------------------------------------------------


function SpeakinSpell:CreateNew_GetSelectedEventObject()
	local eventkey = self.OptionsGUIStates.CreateNew.SelectedEventKey
	if eventkey then
		return SpeakinSpellSavedData.NewEventsDetected[eventkey]
	else
		return nil
	end
end


function SpeakinSpell:CreateNew_ValidateSelectedFilter()
	-- select "show all" if nothing else is selected
	if not self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter then
		self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter = "ALL"
	end

	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end


function SpeakinSpell:CreateNew_ValidateSelectedEvent()
	local DetectedEvent = self:CreateNew_GetSelectedEventObject()

	-- allow the current selection to remain if it is valid
	if DetectedEvent then
		if self:MatchesFilter(DetectedEvent, self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter) then
			-- the current selection is valid, so keep it
			return
		end
		--else, the current selection did not match the filter, so select a default
	end

	-- select the first thing in the list that matches the selected filter, if possible
	for key,DetectedEvent in pairs(SpeakinSpellSavedData.NewEventsDetected) do
		if self:MatchesFilter(DetectedEvent, self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter) then
			-- the current selection is valid, so keep it
			self.OptionsGUIStates.CreateNew.SelectedEventKey = key
			return
		end
	end

	-- our spell list is empty or nothing matches the filter
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
end


function SpeakinSpell:CreateNew_OnEventTypeFilterSelect(getset,value)
	if "GET" == getset then
		--if nothing is selected, set to all
		self:CreateNew_ValidateSelectedFilter()
		return self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter
	else -- "SET"
		-- if the filter is changing, then rebuild the spell/event list
		local rebuild = not (self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter == value)
		-- set the new selection
		self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter = value
		-- rebuild the list to match the new filter
		if rebuild then
			self:CreateNew_RebuildSpellList()
		end
	end
	-- NOTE: the current selection will be repaired to match the new filter in _ValidateSelectedEvent
	--		as a side effect when the GUI framework automatically tries to get the new value of 
	--		the rest of the controls on the page, including the selected event
end


function SpeakinSpell:CreateNew_OnNewSpellNameSelect(getset, value)
	if "GET" == getset then
		--if nothing is selected, try to select the first thing in the list
		self:CreateNew_ValidateSelectedEvent()
		return self.OptionsGUIStates.CreateNew.SelectedEventKey
	else -- "SET"
		self.OptionsGUIStates.CreateNew.SelectedEventKey = value
	end
end


function SpeakinSpell:CreateNew_RebuildSpellList()
	-- apply the event filter while rebuilding the GUI list
	self:CreateNew_ValidateSelectedFilter()
	
	-- reset the list
	self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values = {}
	self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
	
	-- search all new events for those that match the filter
	for key,DetectedEvent in pairs(SpeakinSpellSavedData.NewEventsDetected) do
		-- only use events that match the current filter
		if self:MatchesFilter(DetectedEvent, self.OptionsGUIStates.CreateNew.SelectedEventTypeFilter) then
			-- add to the list of shown values
			self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[key] = DetectedEvent.DisplayName
			-- make sure we have a current spell selected for the options GUI 
			-- we'll use the last one we see because it's easier
			self.OptionsGUIStates.CreateNew.SelectedEventKey = DetectedEvent.key
		end
	end
end

function SpeakinSpell:CreateNew_OnClickCreateNew()
	local funcname = "CreateNew_OnClickCreateNew"
	
	-- grab the selected new event
	local de = self:CreateNew_GetSelectedEventObject()
	if not de then
		self:ErrorMsg(funcname, "no event selected")
		return
	end
	
	-- add the spell to the message table and related data structures
	if not SpeakinSpellSavedData.EventTable[de.key] then
		SpeakinSpellSavedData.EventTable[de.key] = {
			-- event information
			DetectedEvent = {
				key			= de.key,
				name		= de.name,
				rank		= de.rank,
				DisplayName = de.DisplayName,
				SpellLink	= de.SpellLink,
				type		= de.type,
				-- caster	= de.caster -- not stored in event table
				-- target	= de.target -- not stored in event table
				-- pet		= de.pet    -- not stored in event table
			},
			-- default options
			Channels = self:CopyTable(SpeakinSpell.DEFAULTS.CHANNEL_RULES),
			Messages = self:CopyTable(DEFAULT_SPEECHES.NEWSPELL),
			DisableAnnouncements = false,
			WhisperTarget = false,
			Frequency = SpeakinSpell.DEFAULTS.FREQUENCY,
			Cooldown  = SpeakinSpell.DEFAULTS.COOLDOWN,
		}
	end
	
	-- destroy the new event detection since this is now a registered event
	-- NO! keep it.
	-- NOTE: We save the whole list now forever and we don't want to have to recast to get it back
	-- SpeakinSpellSavedData.NewEventsDetected[de.key] = nil
	
	-- deselect it from the Create New screen and remove it from that UI
	-- NO! see note above
	-- self.OptionsGUIStates.CreateNew.SelectedEventKey = nil
	-- self.OptionsGUI.args.CreateNew.args.NewSpellNameSelect.values[de.key] = nil

	-- setup the message settings GUI to show the new category of event we've added
	self.OptionsGUIStates.MessageSettings.SelectedEventTypeFilter = de.type
	-- rebuild the spell list to include our new event, now that we've selected the appropriate filter
	self:CurrentMessagesGUI_RebuildSpellList()
	-- and select our new spell
	self.OptionsGUIStates.MessageSettings.SelectedEventKey = de.key
	
	-- now we're ready to go to the new event's settings in the options GUI
	InterfaceOptionsFrame_OpenToCategory(L["Message Settings TITLE"])
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - HELP PAGES
-------------------------------------------------------------------------------


function SpeakinSpell:HelpPages_Build()
	-- build help pages
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		local NextPageDisplayName = ""
		if i < #(HELPFILE.PAGEORDER) then
			NextPageDisplayName = HELPFILE.PAGEORDER[i+1]
		else --wrap to first page
			NextPageDisplayName = HELPFILE.PAGEORDER[1]
		end
		-- setup OptionGUI object
		self.OptionsGUI.args[FrameObjectName] = {
			order = 3,
			type = "group",
			name = "unused_HelpFrame_name",
			desc = "unused_HelpFrame_desc",
			args = {
				CaptionNormalMode = {
					order = 1,
					type = "description",
					name = HELPFILE.PAGES[ HELPFILE.PAGEORDER[i] ], --help CONTENTS
				},
				NextPage = {
					order = 2,
					type = "execute",
					width = "full",
					name = NextPageDisplayName,
					desc = L["Go to the next help topic"],
					func = function() InterfaceOptionsFrame_OpenToCategory(NextPageDisplayName) end
				}
			}, --end args
		} --end title
		
		-- add a button to the TOC page, linking to the child page we just created above
		self.OptionsGUI.args.HelpRoot.args[FrameObjectName] = {
			order = i+1,
			type = "execute",
			width = "full",
			name = FrameDisplayName,
			desc = FrameDisplayName, --L["Go to the first help topic"], --TODOLATER: tooltips different from labels, and cleanup L[] entry
			func = function() InterfaceOptionsFrame_OpenToCategory(FrameDisplayName) end
		}

	end
end


function SpeakinSpell:HelpPages_AddToBlizOptions()
	--for title,contents in pairs(HELPFILE.PAGES) do --has an unpredictable order
	for i=1,#(HELPFILE.PAGEORDER),1 do
		local FrameObjectName = "HelpTopic"..tostring(i)
		local FrameDisplayName = HELPFILE.PAGEORDER[i]
		--TODOFUTURE: push down to tier 3 -- tier 3 seems buggy, trying tier 2 instead
		--self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell Help"])
		self:AddToBlizOptions(FrameObjectName, FrameDisplayName, L["SpeakinSpell"])
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - SHARED UTILITIES
-------------------------------------------------------------------------------


function SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	-- AddToBlizOptions(appName="SpeakinSpell", display name, parent, ObjectName)
	-- NOTE: the FrameObjectName here represents self.OptionsGUI.args[FrameObjectName]
	self.optionsFrames[FrameObjectName] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpeakinSpell", FrameDisplayName, ParentDisplayName, FrameObjectName)
	
	-- for some reason the framework has a problem if we define the "reset defaults" button handler in the data before calling AddToBlizOptions
	-- so we have to add it afterwards
	self.optionsFrames[FrameObjectName].default = function() self:ResetToDefaults(FrameObjectName) end
end


function SpeakinSpell:InitOptionsGUIState() -- sets up event lists and current selections
	self.OptionsGUIStates = {
		MessageSettings = {
			SelectedEventKey = nil,
			SelectedEventTypeFilter = nil,
		},
		CreateNew = {
			SelectedEventKey = nil,
			SelectedEventTypeFilter = nil,
		}
	}

	-- the options GUI may still have a reference to the old SavedData table.
	-- refresh it since we've rebuilt the table
	-- but if we're initializing then the OptionsGUI may not have been created yet
	if self.OptionsGUI then
		self:CurrentMessagesGUI_RebuildSpellList()
		self:CreateNew_RebuildSpellList()
	end
end


function SpeakinSpell:SetupOptionsGUI_CreateMessageInputs()
	-- add SpeakinSpell.MAX.MESSAGES_PER_SPELL random speech input boxes
	for i=1,SpeakinSpell.MAX.MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.CurrentMessagesGUI.args["SpellMessage"..i] = {
			order = i + 17, -- should be the order of the last item listed above the message input boxes (CooldownSlider), plus i
			width = "full",
			type = "input",
			name = string.format(L["Random Speech i"], i), -- "Random Speech %d"
			desc = L["Random Speech i TOOLTIP"],
			get = function() return  self:CurrentMessagesGUI_OnMessage("GET",i,nil) end,
			set = function(_, value) self:CurrentMessagesGUI_OnMessage("SET",i, value) end,
			guiHidden = true, --by default, until CurrentMessagesGUI_Refresh is called
		}
	end
end



function SpeakinSpell:RefreshFrame(FrameName)
	-- the only way I found to do this is to hide and then show the frame
	-- this occurs fast enough that I do not see a flicker when this happens
	-- UPDATE: saw a forum post when I was at work with the proper API call, but I already forget what it was, and this still works
	if self.optionsFrames and self.optionsFrames[FrameName] and self.optionsFrames[FrameName]:IsVisible() then
		self.optionsFrames[FrameName]:Hide()
		self.optionsFrames[FrameName]:Show()
	end
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS - ENTRY POINT FOR ALL
-------------------------------------------------------------------------------


function SpeakinSpell:SetupOptionsGUI()
	-- Define Options table for the GUI
	self.OptionsGUI = {
		type = "group",
		name = L["SpeakinSpell"],
		args = {
	---------------------------------------------------------------------------------------
			General = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					Caption = {
						order = 1,
						type = "description",
						name = L["General Settings"],
					},
					ShowURL = {
						order = 2,
						type = "description",
						name = SpeakinSpell.URL,
					},
					EnableAllMessagesToggle = {
						order = 3,
						type = "toggle",
						width = "full",
						name = L["Enable All Messages"],
						desc = L["Enable All Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnEnableAllMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnEnableAllMessagesToggle("SET",value) end,
					},
					ShowSpellEventsToggle = {
						order = 4,
						type = "toggle",
						width = "full",
						name = L["Report detected spell events"],
						desc = L["Report detected spell events TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowSpellEventsToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowSpellEventsToggle("SET",value) end,
					},
					ShowDebugMessagesToggle = {
						order = 5,
						type = "toggle",
						width = "full",
						name = L["Show Debugging Messages (verbose)"],
						desc = L["Show Debugging Messages TOOLTIP"],
						get = function() return  self:GeneralOptions_OnShowDebugMessagesToggle("GET",nil) end,
						set = function(_, value) self:GeneralOptions_OnShowDebugMessagesToggle("SET",value) end,
					},
					ImportDefaultsButton = {
						order = 6,
						type = "execute",
						width = "full",
						name = L["Import New Default Sample Speeches"],
						desc = L["Import New Default Sample Speeches TOOLTIP"],
						func = function() self:GeneralOptions_OnClickImport() end,
					},
				}, --end args
			}, --end General

	---------------------------------------------------------------------------------------
			CurrentMessagesGUI = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = L["Message Settings"],
					},
					EventTypeFilterSelect = {
						order = 2,
						type = "select",
						width = "full",
						name = L["Select a Category of Events"],
						desc = L["Select a Category of Events TOOLTIP"],
						values = L.EVENTTYPES.AS_FILTERS,
						get = function() return  self:CurrentMessagesGUI_OnEventTypeFilterSelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnEventTypeFilterSelect("SET",value) end,
					},
					EventSelect = {
						order = 2,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP"],
						values = {},
						get = function() return  self:CurrentMessagesGUI_OnEventSelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnEventSelect("SET",value) end,
					},
					DeleteSpellButton = {
						order = 3,
						type = "execute",
						name = L["Delete Spell"],
						desc = L["Delete Spell TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickDelete() end,
					},
					CreateSpellButton = {
						order = 3,
						type = "execute",
						name = L["Create New..."],
						desc = L["Create New... TOOLTIP"],
						func = function() self:CurrentMessagesGUI_OnClickGotoCreate() end,
					},
					Disable = {
						order = 5,
						type = "toggle",
						width = "full",
						name = L["Disable announcements for this spell"],
						desc = L["Disable announcements for this spell TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnSpellDisable("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnSpellDisable("SET",value) end,
					},
					ChannelSelectCaption = {
						order = 6,
						type = "description",
						width= "full",
						name = L["Select the channels to use when announcing this spell when"],
					},
					ChannelSelectBG = {
						order = 10,
						type = "select",
						name = L["In a Battleground"],
						desc = L["In a Battleground TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","BG",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","BG",value) end,
					},
					ChannelSelectRaid = {
						order = 11,
						type = "select",
						name = L["In a Raid"],
						desc = L["In a Raid TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Raid",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Raid",value) end,
					},
					ChannelSelectParty = {
						order = 12,
						type = "select",
						name = L["In a Party"],
						desc = L["In a Party TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Party",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Party",value) end,
					},
					ChannelSelectSolo = {
						order = 13,
						type = "select",
						name = L["By yourself"],
						desc = L["By yourself TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Solo",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Solo",value) end,
					},
					ChannelSelectArena = {
						order = 14,
						type = "select",
						name = L["In Arena"],
						desc = L["In Arena TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return  self:CurrentMessagesGUI_OnChannelSelect("GET","Arena",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnChannelSelect("SET","Arena",value) end,
					},
					WhisperTargetToggle = {
						order = 15,
						type = "toggle",
						width = "full",
						name = L["Whisper the message to your target"],
						desc = L["Whisper the message to your target TOOLTIP"],
						get = function() return  self:CurrentMessagesGUI_OnWhisperTarget("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnWhisperTarget("SET",value) end,
					},
					FrequencySelect = {
						order = 16,
						type = "range",
						width = "full",
						name = L["How often do you want to announce this spell?"],
						desc = L["How often TOOLTIP"],
						min = 0.01, max = 1.00, step=0.01,
						isPercent = true,
						get = function() return  self:CurrentMessagesGUI_OnFrequencySelect("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnFrequencySelect("SET",value) end,
					},
					CooldownSlider = {
						order = 17,
						type = "range",
						width = "full",
						name = L["Cooldown on announcements for this spell"] ,
						desc = L["Cooldown TOOLTIP"],
						min =0, max = 600, step=1, -- allowing for 0 seconds up to 10 minutes
						get = function() return  self:CurrentMessagesGUI_OnCooldownSlider("GET",nil) end,
						set = function(_, value) self:CurrentMessagesGUI_OnCooldownSlider("SET",value) end,
					},
				}, --end args
			}, --end CurrentMessagesGUI
			
	---------------------------------------------------------------------------------------
			Trade = {
				order = 1,
				type = "group",
				name = "unused_Trade_name",
				desc = "unused_Trade_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = "Coming Soon... randomized trade spam for all of your professional and class services, portals, LFG, LFW",
					},
				}, --end args
			}, --end Trade
			
	---------------------------------------------------------------------------------------
			CreateNew = {
				order = 1,
				type = "group",
				name = "unused_CreateNew_name",
				desc = "unused_CreateNew_desc",
				args = {
					NewSpellModeInstruct = {
						order = 1,
						type = "description",
						name = L["NEW SPELL INSTRUCTIONS"],
					},
					EventTypeFilterSelect = {
						order = 2,
						type = "select",
						width = "full",
						name = L["Select a Category of Events"],
						desc = L["Select a Category of Events TOOLTIP"],
						values = L.EVENTTYPES.AS_FILTERS,
						get = function() return  self:CreateNew_OnEventTypeFilterSelect("GET",nil) end,
						set = function(_, value) self:CreateNew_OnEventTypeFilterSelect("SET",value) end,
					},
					NewSpellNameSelect = {
						order = 3,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP - Create New"],
						values = {},
						get = function() return  self:CreateNew_OnNewSpellNameSelect("GET",nil) end,
						set = function(_, value) self:CreateNew_OnNewSpellNameSelect("SET",value) end,
					},
					CreateNewSpellButton = {
						order = 4,
						type = "execute",
						width = "full",
						name = L["Create Selected Event"],
						desc = L["Create Selected Event TOOLTIP"],
						func = function() self:CreateNew_OnClickCreateNew() end,
					},
				}, --end args
			}, --end CreateNew

	---------------------------------------------------------------------------------------
			HelpRoot = {
				order = 2,
				type = "group",
				name = "unused_HelpRoot_name",
				desc = "unused_HelpRoot_desc",
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = HELPFILE.ROOT,
					},
				}, --end args
			}, --end HelpRoot
			
		}, --end args
	} -- end OptionsGUI
	---------------------------------------------------------------------------------------
	
	-- set up spell/event lists and current selections
	self:InitOptionsGUIState()
	
	-- add SpeakinSpell.MAX.MESSAGES_PER_SPELL random speech input boxes
	self:SetupOptionsGUI_CreateMessageInputs()
	
	-- build help pages
	self:HelpPages_Build()
	
	-- add the options to the in-game interface options panel at Escape > Interface > Addons > SpeakinSpell
	-- The ordering here matters, it determines the order in the Blizzard Interface Options
	-- (or it would matter if we had sub-categories, which we currently don't)
	-- InterfaceOptions_AddCategory(self.OptionsGUI) -- done via Ace below

	-- :RegisterOptionsTable(appName, options, slashcmd, persist)
	-- AceConfig:RegisterOptionsTable("MyAddon", myOptions, {"/myslash", "/my"})
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpeakinSpell", self.OptionsGUI, nil)
	
	--SpeakinSpell:AddToBlizOptions(FrameObjectName, FrameDisplayName, ParentDisplayName)
	--frame objects are stored to self.optionsFrames[FrameObjectName]
	self.optionsFrames = {}
	self:AddToBlizOptions("General",		L["SpeakinSpell"],		nil)
	self:AddToBlizOptions("CurrentMessagesGUI",	L["Message Settings TITLE"],		L["SpeakinSpell"])
	self:AddToBlizOptions("CreateNew",		L["Create New..."],		L["SpeakinSpell"])
	self:AddToBlizOptions("Trade",			L["Trade"],				L["SpeakinSpell"])
	self:AddToBlizOptions("HelpRoot",		L["SpeakinSpell Help"],	L["SpeakinSpell"])
	
	-- add help pages
	self:HelpPages_AddToBlizOptions()

end


